__copyright__ = "Copyright (C) 2018 Kaushik Kulkarni"

__license__ = """
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""

import logging

import numpy as np
import pytest

import pyopencl as cl
import pyopencl.clrandom
from pyopencl.tools import (  # noqa: F401
    pytest_generate_tests_for_pyopencl as pytest_generate_tests,
)
from pytools import ImmutableRecord

import loopy as lp
from loopy.version import LOOPY_USE_LANGUAGE_VERSION_2018_2  # noqa: F401


logger = logging.getLogger(__name__)


def test_register_function_lookup(ctx_factory):
    ctx = ctx_factory()
    queue = cl.CommandQueue(ctx)
    rng = np.random.default_rng(seed=42)

    from testlib import Log2Callable

    x = rng.random(size=10)
    queue = cl.CommandQueue(ctx)

    prog = lp.make_kernel(
            "{[i]: 0<=i<10}",
            """
            y[i] = log2(x[i])
            """)
    prog = lp.register_callable(prog, "log2", Log2Callable("log2"))

    _evt, (out, ) = prog(queue, x=x)

    assert np.linalg.norm(np.log2(x)-out)/np.linalg.norm(np.log2(x)) < 1e-15


@pytest.mark.parametrize("inline", [False, True])
def test_register_knl(ctx_factory, inline):
    ctx = ctx_factory()
    queue = cl.CommandQueue(ctx)
    rng = np.random.default_rng(seed=42)

    n = 4
    x = rng.random(size=(n, n, n, n, n))
    y = rng.random(size=(n, n, n, n, n))

    grandchild_knl = lp.make_function(
            "{[i, j]:0<= i, j< 4}",
            """
            c[i, j] = 2*a[i, j] + 3*b[i, j]
            """, name="linear_combo1")

    child_knl = lp.make_function(
            "{[i, j]:0<=i, j < 4}",
            """
            [i, j]: g[i, j] = linear_combo1([i, j]: e[i, j], [i, j]: f[i, j])
            """, name="linear_combo2")

    parent_knl = lp.make_kernel(
            "{[i, j, k, l, m]: 0<=i, j, k, l, m<4}",
            """
            [j, l]: z[i, j, k, l, m] = linear_combo2([j, l]: x[i, j, k, l, m],
                                                     [j, l]: y[i, j, k, l, m])
            """,
            kernel_data=[
                lp.GlobalArg(
                    name="x, y",
                    dtype=np.float64,
                    shape=(n, n, n, n, n)),
                ...]
            )

    knl = lp.merge([grandchild_knl, child_knl, parent_knl])

    if inline:
        knl = lp.inline_callable_kernel(knl, "linear_combo2")
        knl = lp.inline_callable_kernel(knl, "linear_combo1")

    _evt, (out, ) = knl(queue, x=x, y=y)

    assert (np.linalg.norm(2*x+3*y-out)/(
        np.linalg.norm(2*x+3*y))) < 1e-15


@pytest.mark.parametrize("inline", [False, True])
def test_slices_with_negative_step(ctx_factory, inline):
    ctx = ctx_factory()
    queue = cl.CommandQueue(ctx)
    rng = np.random.default_rng(seed=42)

    n = 4
    x = rng.random(size=(n, n, n, n, n))
    y = rng.random(size=(n, n, n, n, n))

    child_knl = lp.make_function(
            "{[i, j]:0<=i, j < 4}",
            """
            g[i, j] = 2*e[i, j] + 3*f[i, j]
            """, name="linear_combo")

    parent_knl = lp.make_kernel(
            "{[i, k, m]: 0<=i, k, m<4}",
            """
            z[i, 3:-1:-1, k, :, m] = linear_combo(x[i, :, k, :, m],
                                                   y[i, :, k, :, m])
            """,
            kernel_data=[
                lp.GlobalArg(
                    name="x, y, z",
                    dtype=np.float64,
                    shape=(n, n, n, n, n)),
                ...]
            )

    knl = lp.merge([parent_knl, child_knl])
    if inline:
        knl = lp.inline_callable_kernel(knl, "linear_combo")

    _evt, (out, ) = knl(queue, x=x, y=y)

    assert (np.linalg.norm(2*x+3*y-out[:, ::-1, :, :, :])/(
        np.linalg.norm(2*x+3*y))) < 1e-15


@pytest.mark.parametrize("inline", [False, True])
def test_register_knl_with_hw_axes(ctx_factory, inline):
    ctx = ctx_factory()
    queue = cl.CommandQueue(ctx)

    n = 4

    x_dev = cl.clrandom.rand(queue, (n, n, n, n, n), np.float64)
    y_dev = cl.clrandom.rand(queue, (n, n, n, n, n), np.float64)

    callee_knl = lp.make_function(
            "{[i, j]:0<=i, j < 4}",
            """
            g[i, j] = 2*e[i, j] + 3*f[i, j]
            """, name="linear_combo")

    callee_knl = lp.split_iname(callee_knl, "i", 1, inner_tag="l.0", outer_tag="g.0")

    caller_knl = lp.make_kernel(
            "{[i, j, k, l, m]: 0<=i, j, k, l, m<4}",
            """
            [j, l]: z[i, j, k, l, m] = linear_combo([j, l]: x[i, j, k, l, m],
                                                     [j, l]: y[i, j, k, l, m])
            """, name="caller")
    caller_knl = lp.split_iname(caller_knl, "i", 4, inner_tag="l.1", outer_tag="g.1")

    knl = lp.merge([caller_knl, callee_knl])

    knl = lp.set_options(knl, return_dict=True)

    if inline:
        knl = lp.inline_callable_kernel(knl, "linear_combo")

    _evt, out = knl(queue, x=x_dev, y=y_dev)

    x_host = x_dev.get()
    y_host = y_dev.get()

    assert np.linalg.norm(2*x_host+3*y_host-out["z"].get())/np.linalg.norm(
            2*x_host+3*y_host) < 1e-15


@pytest.mark.parametrize("inline", [False, True])
def test_shape_translation_through_sub_array_ref(ctx_factory, inline):
    ctx = ctx_factory()
    queue = cl.CommandQueue(ctx)

    x1 = cl.clrandom.rand(queue, (3, 2), dtype=np.float64)
    x2 = cl.clrandom.rand(queue, (6, ), dtype=np.float64)
    x3 = cl.clrandom.rand(queue, (6, 6), dtype=np.float64)

    callee1 = lp.make_function(
            "{[i]: 0<=i<6}",
            """
            b[i] = 2*abs(a[i])
            """, name="callee_fn1")

    callee2 = lp.make_function(
            "{[i, j]: 0<=i<3 and 0 <= j < 2}",
            """
            b[i, j] = 3*a[i, j]
            """, name="callee_fn2")

    callee3 = lp.make_function(
            "{[i]: 0<=i<6}",
            """
            b[i] = 5*a[i]
            """, name="callee_fn3")

    knl = lp.make_kernel(
            "{[i, j, k, l]:  0<= i < 6 and 0 <= j < 3 and 0 <= k < 2 and 0<=l<6}",
            """
            [i]: y1[i//2, i%2] = callee_fn1([i]: x1[i//2, i%2])
            [j, k]: y2[2*j+k] = callee_fn2([j, k]: x2[2*j+k])
            [l]: y3[l, l] = callee_fn3([l]: x3[l, l])
            """)

    knl = lp.merge([knl, callee1])
    knl = lp.merge([knl, callee2])
    knl = lp.merge([knl, callee3])

    if inline:
        knl = lp.inline_callable_kernel(knl, "callee_fn1")
        knl = lp.inline_callable_kernel(knl, "callee_fn2")
        knl = lp.inline_callable_kernel(knl, "callee_fn3")

    knl = lp.set_options(knl, write_code=True)
    knl = lp.set_options(knl, return_dict=True)
    _evt, out_dict = knl(queue, x1=x1, x2=x2, x3=x3)

    y1 = out_dict["y1"].get()
    y2 = out_dict["y2"].get()
    y3 = out_dict["y3"].get()

    assert (np.linalg.norm(y1-2*x1.get())) < 1e-15
    assert (np.linalg.norm(y2-3*x2.get())) < 1e-15
    assert (np.linalg.norm(np.diag(y3-5*x3.get()))) < 1e-15


def test_multi_arg_array_call(ctx_factory):
    ctx = ctx_factory()
    queue = cl.CommandQueue(ctx)
    rng = np.random.default_rng(seed=42)

    import pymbolic.primitives as p

    n = 10
    acc_i = p.Variable("acc_i")
    i = p.Variable("i")
    index = p.Variable("index")
    a_i = p.Subscript(p.Variable("a"), p.Variable("i"))

    argmin_kernel = lp.make_function(
            "{[i]: 0 <= i < n}",
            [
                lp.Assignment(id="init2", assignee=index,
                    expression=0),
                lp.Assignment(id="init1", assignee=acc_i,
                    expression="214748367"),
                lp.Assignment(id="insn", assignee=index,
                    expression=p.If(p.Expression.eq(acc_i, a_i), i, index),
                    depends_on="update"),
                lp.Assignment(id="update", assignee=acc_i,
                    expression=p.Variable("min")(acc_i, a_i),
                    depends_on="init1,init2")],
            [
                lp.GlobalArg("a"),
                lp.GlobalArg("acc_i, index", is_input=False, is_output=True,
                             shape=lp.auto),
                ...],
            name="custom_argmin")

    argmin_kernel = lp.fix_parameters(argmin_kernel, n=n)

    knl = lp.make_kernel(
            "{[i]:0<=i<n}",
            """
            []: min_val[()], []: min_index[()] = custom_argmin([i]:b[i])
            """)

    knl = lp.fix_parameters(knl, n=n)
    knl = lp.set_options(knl, return_dict=True)
    knl = lp.merge([knl, argmin_kernel])

    b = rng.normal(size=n)
    _evt, out_dict = knl(queue, b=b)
    tol = 1e-15
    from numpy.linalg import norm
    assert norm(out_dict["min_val"] - np.min(b)) < tol
    assert norm(out_dict["min_index"] - np.argmin(b)) < tol


@pytest.mark.parametrize("inline", [False, True])
def test_packing_unpacking(ctx_factory, inline):
    ctx = ctx_factory()
    queue = cl.CommandQueue(ctx)

    x1 = cl.clrandom.rand(queue, (3, 2), dtype=np.float64)
    x2 = cl.clrandom.rand(queue, (6, ), dtype=np.float64)

    callee1 = lp.make_function(
            "{[i]: 0<=i<6}",
            """
            b[i] = 2*a[i]
            """, name="callee_fn1")

    callee2 = lp.make_function(
            "{[i, j]: 0<=i<2 and 0 <= j < 3}",
            """
            b[i, j] = 3*a[i, j]
            """, name="callee_fn2")

    knl = lp.make_kernel(
            "{[i, j, k]:  0<= i < 3 and 0 <= j < 2 and 0 <= k < 6}",
            """
            [i, j]: y1[i, j] = callee_fn1([i, j]: x1[i, j])
            [k]: y2[k] = callee_fn2([k]: x2[k])
            """)

    knl = lp.merge([knl, callee1])
    knl = lp.merge([knl, callee2])

    knl = lp.pack_and_unpack_args_for_call(knl, "callee_fn1")
    knl = lp.pack_and_unpack_args_for_call(knl, "callee_fn2")

    if inline:
        knl = lp.inline_callable_kernel(knl, "callee_fn1")
        knl = lp.inline_callable_kernel(knl, "callee_fn2")

    knl = lp.set_options(knl, write_code=True)
    knl = lp.set_options(knl, return_dict=True)
    _evt, out_dict = knl(queue, x1=x1, x2=x2)

    y1 = out_dict["y1"].get()
    y2 = out_dict["y2"].get()

    assert np.linalg.norm(2*x1.get()-y1)/np.linalg.norm(
            2*x1.get()) < 1e-15
    assert np.linalg.norm(3*x2.get()-y2)/np.linalg.norm(
            3*x2.get()) < 1e-15


@pytest.mark.parametrize("inline", [False, True])
def test_empty_sub_array_refs(ctx_factory, inline):
    # See: https://github.com/OP2/PyOP2/pull/559#discussion_r272208618
    ctx = ctx_factory()
    queue = cl.CommandQueue(ctx)
    rng = np.random.default_rng(seed=42)

    x = rng.normal(size=10)
    y = rng.normal(size=10)

    callee = lp.make_function(
            "{[d]:0<=d<1}",
            """
            c[d] = a[d] - b[d]
            """, name="wence_function")

    caller = lp.make_kernel("{[i,k]: 0<=i<10 and 0<=k<1}",
            """
            [k]:z[i+k] = wence_function([k]:x[i+k], [k]:y[i+k])
            """,
            [lp.GlobalArg("x, y", dtype=np.float64, shape=(10, )), ...])

    caller = lp.merge([caller, callee])

    if inline:
        caller = lp.inline_callable_kernel(caller, "wence_function")

    _evt, (out, ) = caller(queue, x=x, y=y)
    assert np.allclose(out, x-y)


@pytest.mark.parametrize("inline", [False, True])
def test_array_inputs_to_callee_kernels(ctx_factory, inline):
    ctx = ctx_factory()
    queue = cl.CommandQueue(ctx)
    rng = np.random.default_rng(seed=42)

    n = 2 ** 3
    x = rng.normal(size=(n, n))
    y = rng.normal(size=(n, n))

    child_knl = lp.make_function(
            "{[i, j]:0<=i, j < 8}",
            """
            g[i, j] = 2*e[i, j] + 3*f[i, j]
            """, name="linear_combo")

    parent_knl = lp.make_kernel(
            "{:}",
            """
            z[:, :] = linear_combo(x, y)
            """,
            kernel_data=[
                lp.GlobalArg(
                    name="x, y, z",
                    dtype=np.float64,
                    shape=(n, n)),
                ...]
            )

    knl = lp.merge([parent_knl, child_knl])
    if inline:
        knl = lp.inline_callable_kernel(knl, "linear_combo")

    _evt, (out, ) = knl(queue, x=x, y=y)

    assert (np.linalg.norm(2*x+3*y-out)/(
        np.linalg.norm(2*x+3*y))) < 1e-15


def test_stride_depending_on_args(ctx_factory):
    ctx = ctx_factory()

    twice = lp.make_function(
            "{[i, j]: 0<=i, j < n}",
            """
            b[i, j] = 2*a[i, j]
            """, [lp.ValueArg("n"), lp.GlobalArg("a"), lp.GlobalArg("b")],
            name="twice")

    thrice = lp.make_function(
            "{[i, j]: 0<=i, j < n}",
            """
            b[i, j] = 3*a[i, j]
            """, [lp.ValueArg("n"), lp.GlobalArg("a", shape=lp.auto),
                lp.GlobalArg("b", shape=lp.auto)],
            name="thrice")

    prog = lp.make_kernel(
            "{[i0,i1,i2,i3,i4,i5,i6,i7]: 0<=i0, i1, i2, i3, i4, i5, i6, i7< N}",
            """
            [i0, i1]: y[i0, i1] = twice(N, [i2, i3]: x[2*i2, i3])
            [i4, i5]: z[i4, i5] = thrice(N, [i6, i7]: x[2*i6+1, i7])
            """, [
                lp.ValueArg("N", dtype=np.int32), lp.GlobalArg("x",
                    shape=lp.auto, dtype=np.float64), ...])

    prog = lp.merge([prog, twice])
    prog = lp.merge([prog, thrice])

    lp.auto_test_vs_ref(prog, ctx, prog, parameters={"N": 4})


def test_unknown_stride_to_callee(ctx_factory):
    ctx = ctx_factory()

    twice = lp.make_function(
            "{[i, j]: 0<=i, j < n}",
            """
            b[i, j] = 2*a[i, j]
            """, [lp.ValueArg("n"), lp.GlobalArg("a"), lp.GlobalArg("b")],
            name="twice")

    prog = lp.make_kernel(
            "{[i,i0,i1,i2,i3]: 0<=i0, i1, i2, i3< N and 0<=i<Nvar}",
            """
            [i0, i1]: y[i0, i1, i] = twice(N, [i2, i3]: x[2*i2, i3, i])
            """, [
                lp.ValueArg("N", dtype=np.int32), lp.ValueArg("Nvar",
                    dtype=np.int32), lp.GlobalArg("x", shape=lp.auto,
                        dtype=np.float64), ...])

    prog = lp.merge([prog, twice])

    lp.auto_test_vs_ref(prog, ctx, prog, parameters={"N": 4, "Nvar": 5})


def test_argument_matching_for_inplace_update(ctx_factory):
    ctx = ctx_factory()
    queue = cl.CommandQueue(ctx)
    rng = np.random.default_rng(seed=42)

    twice = lp.make_function(
            "{[i]: 0<=i<10}",
            """
            x[i] = 2*x[i]
            """, name="twice")

    knl = lp.make_kernel(
            "{:}",
            """
            x[:] = twice(x[:])
            """, [lp.GlobalArg("x", shape=(10,), dtype=np.float64)])

    knl = lp.merge([knl, twice])

    x = rng.normal(size=10)
    _evt, (out, ) = knl(queue, x=np.copy(x))

    assert np.allclose(2*x, out)


def test_non_zero_start_in_subarray_ref(ctx_factory):
    ctx = ctx_factory()
    queue = cl.CommandQueue(ctx)
    rng = np.random.default_rng(seed=42)

    twice = lp.make_function(
            "{[i]: 0<=i<10}",
            """
            b[i] = 2*a[i]
            """, name="twice")

    knl = lp.make_kernel(
            "{[i, j]: -5<=i<5 and 0<=j<10}",
            """
            [i]:y[i+5] = twice([j]: x[j])
            """, [lp.GlobalArg("x, y", shape=(10,), dtype=np.float64)])
    knl = lp.merge([knl, twice])

    x = rng.normal(size=10)
    _evt, (out, ) = knl(queue, x=np.copy(x))

    assert np.allclose(2*x, out)


def test_incomplete_entrypoint_raises_type_inf_failure():
    from loopy.diagnostic import LoopyError

    twice = lp.make_kernel(
            "{[i]: 0<=i<10}",
            """
            y[i] = 2*x[i]
            """, name="dosify")

    quadr = lp.make_kernel(
            "{:}",
            """
            y[:] = dosify(x[:])
            y[:] = dosify(y[:])
            """, [lp.GlobalArg("x,y", shape=(10,))], name="cuatroify",
            seq_dependencies=True)

    prog = lp.merge([quadr, twice])

    with pytest.raises(LoopyError):
        # 'twice' is also registered as an entrypoint but provided args aren't
        # enough to infer the types
        lp.generate_code_v2(prog)


def test_callees_with_gbarriers_are_inlined(ctx_factory):
    ctx = ctx_factory()
    queue = cl.CommandQueue(ctx)

    ones_and_zeros = lp.make_function(
            "{[i, j]: 0<=i<6 and 0<=j<3}",
            """
            x[i] = 0.0f
            ...gbarrier
            x[j] = 1.0f
            """,
            seq_dependencies=True,
            name="ones_and_zeros")

    t_unit = lp.make_kernel(
            "{ : }",
            """
            y[:] = ones_and_zeros()
            """, [lp.GlobalArg("y", shape=6, dtype=None)])

    t_unit = lp.merge([t_unit, ones_and_zeros])
    _evt, (out,) = t_unit(queue)

    expected_out = np.array([1, 1, 1, 0, 0, 0]).astype(np.float32)

    assert (expected_out == out.get()).all()


def test_callees_with_gbarriers_are_inlined_with_nested_calls(ctx_factory):
    ctx = ctx_factory()
    queue = cl.CommandQueue(ctx)

    ones_and_zeros = lp.make_function(
            "{[i, j]: 0<=i<6 and 0<=j<3}",
            """
            x[i] = 0.0f
            ...gbarrier
            x[j] = 1.0f
            """,
            seq_dependencies=True,
            name="ones_and_zeros")

    dummy_ones_and_zeros = lp.make_function(
            "{[i]: 0<=i<6}",
            """
            [i]: y[i] = ones_and_zeros()
            """,
            name="dummy_ones_and_zeros")

    t_unit = lp.make_kernel(
            "{ : }",
            """
            y[:] = dummy_ones_and_zeros()
            """, [lp.GlobalArg("y", shape=6, dtype=None)])

    t_unit = lp.merge([t_unit, dummy_ones_and_zeros, ones_and_zeros])
    _evt, (out,) = t_unit(queue)

    expected_out = np.array([1, 1, 1, 0, 0, 0]).astype(np.float32)

    assert (expected_out == out.get()).all()


def test_inlining_with_indirections(ctx_factory):
    ctx = ctx_factory()
    queue = cl.CommandQueue(ctx)

    ones_and_zeros = lp.make_function(
            "{[i, j]: 0<=i<6 and 0<=j<3}",
            """
            x[i] = 0.0f
            ...gbarrier
            x[map[j]] = 1.0f
            """,
            seq_dependencies=True,
            name="ones_and_zeros")

    t_unit = lp.make_kernel(
            "{ : }",
            """
            y[:] = ones_and_zeros(mymap[:])
            """, [lp.GlobalArg("y", shape=6, dtype=None),
                  lp.GlobalArg("mymap", dtype=np.int32, shape=3)])

    t_unit = lp.merge([t_unit, ones_and_zeros])
    t_unit = lp.inline_callable_kernel(t_unit, "ones_and_zeros")

    map_in = np.arange(3).astype(np.int32)

    _evt, (out, ) = t_unit(queue, mymap=map_in)

    expected_out = np.array([1, 1, 1, 0, 0, 0]).astype(np.float32)
    assert (expected_out == out).all()


def test_inlining_with_callee_domain_param(ctx_factory):
    ctx = ctx_factory()
    queue = cl.CommandQueue(ctx)

    fill2 = lp.make_function(
            "{[i]: 0<=i<n}",
            """
            y[i] = 2.0
            """,
            name="fill2")

    caller = lp.make_kernel(
            "{[i]: 0<=i<10}",
            """
            [i]: res[i] = fill2(10)
            """)

    caller = lp.merge([caller, fill2])
    caller = lp.inline_callable_kernel(caller, "fill2")
    _evt, (out, ) = caller(queue)

    assert (out == 2).all()


def test_double_resolving():
    from loopy.kernel import KernelState
    from loopy.symbolic import ResolvedFunction
    from loopy.translation_unit import resolve_callables

    knl = lp.make_kernel(
            "{[i]: 0<=i<10}",
            """
            y[i] = sin(x[i])
            """,
            [
                lp.GlobalArg("x", dtype=float, shape=lp.auto),
                ...],
            name="foo"
            )

    knl = resolve_callables(knl)
    knl = knl.with_kernel(knl["foo"].copy(state=KernelState.INITIAL))
    knl = resolve_callables(knl)

    assert "sin" in knl.callables_table
    assert isinstance(knl["foo"].instructions[0].expression.function,
                      ResolvedFunction)


@pytest.mark.parametrize("inline", [False, True])
def test_passing_and_getting_scalar_in_clbl_knl(ctx_factory, inline):
    ctx = cl.create_some_context()
    cq = cl.CommandQueue(ctx)

    call_sin = lp.make_function(
        "{:}",
        """
        y = sin(x)
        """, name="call_sin")

    knl = lp.make_kernel(
        "{:}",
        """
        []: real_y[()] = call_sin(real_x)
        """)

    knl = lp.merge([knl, call_sin])
    knl = lp.set_options(knl, write_code=True)
    if inline:
        knl = lp.inline_callable_kernel(knl, "call_sin")

    _evt, (_out,) = knl(cq, real_x=np.asarray(3.0, dtype=float))


@pytest.mark.parametrize("inline", [False, True])
def test_passing_scalar_as_indexed_subcript_in_clbl_knl(ctx_factory, inline):
    ctx = cl.create_some_context()
    cq = cl.CommandQueue(ctx)
    rng = np.random.default_rng(seed=42)

    twice = lp.make_function(
        "{ : }",
        """
        y = 2*x
        """,
        name="twice")

    knl = lp.make_kernel(
        "{ : }",
        """
        []: Y[0, 0] = twice(X)
        """)

    knl = lp.add_dtypes(knl, {"X": np.float64})
    knl = lp.merge([knl, twice])

    if inline:
        knl = lp.inline_callable_kernel(knl, "twice")

    x_in = rng.random()
    _evt, (out,) = knl(cq, X=x_in)

    np.testing.assert_allclose(out.get(), 2*x_in)


def test_symbol_mangler_in_call(ctx_factory):
    from library_for_test import preamble_for_x, symbol_x
    ctx = cl.create_some_context()
    cq = cl.CommandQueue(ctx)

    knl = lp.make_kernel(
        "{:}",
        """
        y = sin(X)
        """,
        [lp.GlobalArg("y", shape=lp.auto)])

    knl = lp.register_symbol_manglers(knl, [symbol_x])

    knl = lp.register_preamble_generators(knl, [preamble_for_x])

    _evt, (out,) = knl(cq)
    np.testing.assert_allclose(out.get(), np.sin(10))


@pytest.mark.parametrize("which", ["max", "min"])
def test_int_max_min_c_target(ctx_factory, which):
    from numpy.random import default_rng

    from pymbolic import parse
    rng = default_rng()

    n = 100
    arr1 = rng.integers(-100, 100, n)
    arr2 = rng.integers(-100, 100, n)
    np_func = getattr(np, f"{which}imum")

    knl = lp.make_kernel(
        "{[i]: 0<=i<100}",
        [lp.Assignment(parse("out[i]"),
                       parse(f"{which}(arr1[i], arr2[i])"))],
        target=lp.ExecutableCTarget())

    _, (out,) = knl(arr1=arr1, arr2=arr2)
    np.testing.assert_allclose(np_func(arr1, arr2), out)


def test_valueargs_being_mapped_in_inling(ctx_factory):
    doublify = lp.make_function(
            "{[i]: 0<=i<n}",
            """
            y[i] = n*x[i]
            """,
            [lp.ValueArg("n", dtype=np.int32), ...],
            name="doublify",
            )

    knl = lp.make_kernel(
            "{[i, j]: 0<=i, j<10}",
            """
            [i]: bar[i] = doublify(10, [j]: foo[j])
            """,
            [
                lp.GlobalArg("foo", dtype=float, shape=lp.auto),
                ...],
            )
    knl = lp.merge([knl, doublify])
    knl = lp.inline_callable_kernel(knl, "doublify")

    lp.auto_test_vs_ref(knl, ctx_factory(), knl)


@pytest.mark.parametrize("inline", [True, False])
def test_unused_hw_axes_in_callee(ctx_factory, inline):
    ctx = ctx_factory()

    twice = lp.make_function(
            "{[i]: 0<=i<10}",
            """
            y[i] = 2*x[i]
            """, name="twice")

    knl = lp.make_kernel(
            "{[i]: 0<=i<10}",
            """
            y[:, i] = twice(x[:, i])
            """, [lp.GlobalArg("x", shape=(10, 10), dtype=float),
                lp.GlobalArg("y", shape=(10, 10))],
            name="outer")

    twice = lp.tag_inames(twice, {"i": "l.1"})
    knl = lp.tag_inames(knl, {"i": "l.0"})
    knl = lp.merge([knl, twice])

    if inline:
        knl = lp.inline_callable_kernel(knl, "twice")

    lp.auto_test_vs_ref(knl, ctx, knl)


@pytest.mark.parametrize("inline", [True, False])
def test_double_hw_axes_used_in_knl_call(inline):
    from loopy.diagnostic import LoopyError

    twice = lp.make_function(
            "{[i]: 0<=i<10}",
            """
            y[i] = 2*x[i]
            """, name="twice")

    knl = lp.make_kernel(
            "{[i]: 0<=i<10}",
            """
            y[:, i] = twice(x[:, i])
            """, [lp.GlobalArg("x", shape=(10, 10), dtype=float),
                lp.GlobalArg("y", shape=(10, 10))],
            name="outer")

    twice = lp.tag_inames(twice, {"i": "l.0"})
    knl = lp.tag_inames(knl, {"i": "l.0"})
    knl = lp.merge([knl, twice])

    if inline:
        knl = lp.inline_callable_kernel(knl, "twice")

    with pytest.raises(LoopyError):
        lp.generate_code_v2(knl)


@pytest.mark.parametrize("inline", [True, False])
def test_kc_with_floor_div_in_expr(ctx_factory, inline):
    # See https://github.com/inducer/loopy/issues/366
    import loopy as lp

    ctx = ctx_factory()
    callee = lp.make_function(
            "{[i]: 0<=i<10}",
            """
            x[i] = 2*x[i]
            """, name="callee_with_update")

    knl = lp.make_kernel(
            "{[i]: 0<=i<10}",
            """
            [i]: x[2*(i//2) + (i%2)] = callee_with_update([i]: x[i])
            """)

    knl = lp.merge([knl, callee])

    if inline:
        knl = lp.inline_callable_kernel(knl, "callee_with_update")

    lp.auto_test_vs_ref(knl, ctx, knl)


@pytest.mark.parametrize("start", [5, 6, 7])
@pytest.mark.parametrize("inline", [True, False])
def test_non1_step_slices(ctx_factory, start, inline):
    # See https://github.com/inducer/loopy/pull/222#discussion_r645905188

    ctx = ctx_factory()
    cq = cl.CommandQueue(ctx)

    callee = lp.make_function(
            "{[i]: 0<=i<n}",
            """
            y[i] = i**2
            """,
            [lp.ValueArg("n"), ...],
            name="squared_arange")

    t_unit = lp.make_kernel(
            "{[i_init, j_init]: 0<=i_init, j_init<40}",
            f"""
            X[i_init] = 42
            X[{start}:40:3] = squared_arange({len(range(start, 40, 3))})

            Y[j_init] = 1729
            Y[39:{start}:-3] = squared_arange({len(range(39, start, -3))})
            """,
            [lp.GlobalArg("X,Y", shape=40)],
            seq_dependencies=True)

    expected_out1 = 42*np.ones(40, dtype=np.int64)
    expected_out1[start:40:3] = np.arange(len(range(start, 40, 3)))**2

    expected_out2 = 1729*np.ones(40, dtype=np.int64)
    expected_out2[39:start:-3] = np.arange(len(range(39, start, -3)))**2

    t_unit = lp.merge([t_unit, callee])

    t_unit = lp.set_options(t_unit, return_dict=True)

    if inline:
        t_unit = lp.inline_callable_kernel(t_unit, "squared_arange")

    _evt, out_dict = t_unit(cq)

    np.testing.assert_allclose(out_dict["X"].get(), expected_out1)
    np.testing.assert_allclose(out_dict["Y"].get(), expected_out2)


def test_check_bounds_with_caller_assumptions(ctx_factory):
    import islpy as isl

    from loopy.diagnostic import LoopyIndexError

    arange = lp.make_function(
        "{[i]: 0<=i<n}",
        """
        y[i] = i
        """, name="arange")

    knl = lp.make_kernel(
        "{[i]: 0<=i<20}",
        """
        [i]: Y[i] = arange(N)
        """,
        [lp.GlobalArg("Y", shape=(20,)), lp.ValueArg("N", dtype=np.int32)],
        name="epoint")

    knl = lp.merge([knl, arange])

    with pytest.raises(LoopyIndexError):
        lp.generate_code_v2(knl)

    knl = knl.with_kernel(lp.assume(knl.default_entrypoint,
                                    isl.BasicSet("[N] -> { : N <= 20}")))

    lp.auto_test_vs_ref(knl, ctx_factory(),
                        parameters={"N": 15})


def test_callee_with_auto_offset(ctx_factory):

    ctx = ctx_factory()
    queue = cl.CommandQueue(ctx)

    arange = lp.make_function(
        "{[i]: 0<=i<7}",
        """
        y[i] = 2*y[i]
        """,
        [lp.GlobalArg("y",  offset=lp.auto)],
        name="dosify")

    knl = lp.make_kernel(
        "{[i]: 0<=i<7}",
        """
        [i]: y[i] = dosify([i]: y[i])
        """,
        [lp.GlobalArg("y", offset=3, shape=10)])

    knl = lp.merge([knl, arange])

    y = np.arange(10)
    knl(queue, y=y)
    np.testing.assert_allclose(y[:3], np.arange(3))
    np.testing.assert_allclose(y[3:], 2*np.arange(3, 10))


def test_callee_with_parameter_and_grid(ctx_factory):
    ctx = ctx_factory()
    cq = cl.CommandQueue(ctx)

    callee = lp.make_function(
        "{[i]: 0<=i<n}",
        """
        y[i] = i
        """, name="arange")

    knl = lp.make_kernel(
        "{[i]: 0<=i<10}",
        """
        [i]: y[i] = arange(10)
        """)

    knl = lp.merge([callee, knl])
    knl = lp.split_iname(knl, "i", 2,
                         outer_tag="g.0", within="in_kernel:arange")

    _evt, (out,) = knl(cq)
    np.testing.assert_allclose(out.get(), np.arange(10))


@pytest.mark.parametrize("inline", [True, False])
def test_inlining_does_not_require_barrier(inline):

    # {{{ provided by isuruf in https://github.com/inducer/loopy/issues/578

    fft_knl = lp.make_function(
        [
            "{ [i] : 0 <= i <= 1322 }",
            "{ [ifft_0] : 0 <= ifft_0 <= 188 }",
            "{ [iN1_0] : 0 <= iN1_0 <= 6 }",
            "{ [iN1_sum_0] : 0 <= iN1_sum_0 <= 6 }",
            "{ [iN2_0] : iN2_0 = 0 }",
            "{ [i_0] : 0 <= i_0 <= 1322 }",
            "{ [i2_0] : 0 <= i2_0 <= 1322 }",
            "{ [ifft_1] : 0 <= ifft_1 <= 26 }",
            "{ [iN1_1] : 0 <= iN1_1 <= 6 }",
            "{ [iN1_sum_1] : 0 <= iN1_sum_1 <= 6 }",
            "{ [iN2_1] : 0 <= iN2_1 <= 6 }",
            "{ [i_1] : 0 <= i_1 <= 1322 }",
            "{ [i2_1] : 0 <= i2_1 <= 1322 }",
            "{ [ifft_2] : 0 <= ifft_2 <= 8 }",
            "{ [iN1_2] : 0 <= iN1_2 <= 2 }",
            "{ [iN1_sum_2] : 0 <= iN1_sum_2 <= 2 }",
            "{ [iN2_2] : 0 <= iN2_2 <= 48 }",
            "{ [i_2] : 0 <= i_2 <= 1322 }",
            "{ [i2_2] : 0 <= i2_2 <= 1322 }",
            "{ [ifft_3] : 0 <= ifft_3 <= 2 }",
            "{ [iN1_3] : 0 <= iN1_3 <= 2 }",
            "{ [iN1_sum_3] : 0 <= iN1_sum_3 <= 2 }",
            "{ [iN2_3] : 0 <= iN2_3 <= 146 }",
            "{ [i_3] : 0 <= i_3 <= 1322 }",
            "{ [i2_3] : 0 <= i2_3 <= 1322 }",
            "{ [ifft_4] : ifft_4 = 0 }",
            "{ [iN1_4] : 0 <= iN1_4 <= 2 }",
            "{ [iN1_sum_4] : 0 <= iN1_sum_4 <= 2 }",
            "{ [iN2_4] : 0 <= iN2_4 <= 440 }",
            "{ [i_4] : 0 <= i_4 <= 1322 }",
            "{ [i2_4] : 0 <= i2_4 <= 1322 }",
        ],
        """
        exp_table[i] = exp((-0.004749195243521985j)*i) {id=exp_table}
        temp[i_0] = x[i_0] {id=copy_0, dep=exp_table}
        x[i2_0] = 0 {id=reset_0, dep=copy_0}
        table_idx_0 = 189*iN1_sum_0*(iN2_0 + iN1_0) {id=idx_0, dep=reset_0}
        exp_0 = exp_table[table_idx_0 % 1323] {id=exp_0, dep=idx_0:exp_table}
        x[ifft_0 + 189*(iN1_0 + iN2_0)] = x[ifft_0 + 189*(iN1_0 + iN2_0)] + exp_0*temp[ifft_0 + 189*(iN2_0*7 + iN1_sum_0)] {id=update_0, dep=exp_0}
        temp[i_1] = x[i_1] {id=copy_1, dep=update_0}
        x[i2_1] = 0 {id=reset_1, dep=copy_1}
        table_idx_1 = 27*iN1_sum_1*(iN2_1 + 7*iN1_1) {id=idx_1, dep=reset_1}
        exp_1 = exp_table[table_idx_1 % 1323] {id=exp_1, dep=idx_1:exp_table}
        x[ifft_1 + 27*(iN1_1*7 + iN2_1)] = x[ifft_1 + 27*(iN1_1*7 + iN2_1)] + exp_1*temp[ifft_1 + 27*(iN2_1*7 + iN1_sum_1)] {id=update_1, dep=exp_1}
        temp[i_2] = x[i_2] {id=copy_2, dep=update_1}
        x[i2_2] = 0 {id=reset_2, dep=copy_2}
        table_idx_2 = 9*iN1_sum_2*(iN2_2 + 49*iN1_2) {id=idx_2, dep=reset_2}
        exp_2 = exp_table[table_idx_2 % 1323] {id=exp_2, dep=exp_table:idx_2}
        x[ifft_2 + 9*(iN1_2*49 + iN2_2)] = x[ifft_2 + 9*(iN1_2*49 + iN2_2)] + exp_2*temp[ifft_2 + 9*(iN2_2*3 + iN1_sum_2)] {id=update_2, dep=exp_2}
        temp[i_3] = x[i_3] {id=copy_3, dep=update_2}
        x[i2_3] = 0 {id=reset_3, dep=copy_3}
        table_idx_3 = 3*iN1_sum_3*(iN2_3 + 147*iN1_3) {id=idx_3, dep=reset_3}
        exp_3 = exp_table[table_idx_3 % 1323] {id=exp_3, dep=exp_table:idx_3}
        x[ifft_3 + 3*(iN1_3*147 + iN2_3)] = x[ifft_3 + 3*(iN1_3*147 + iN2_3)] + exp_3*temp[ifft_3 + 3*(iN2_3*3 + iN1_sum_3)] {id=update_3, dep=exp_3}
        temp[i_4] = x[i_4] {id=copy_4, dep=update_3}
        x[i2_4] = 0 {id=reset_4, dep=copy_4}
        table_idx_4 = iN1_sum_4*(iN2_4 + 441*iN1_4) {id=idx_4, dep=reset_4}
        exp_4 = exp_table[table_idx_4 % 1323] {id=exp_4, dep=exp_table:idx_4}
        x[ifft_4 + iN1_4*441 + iN2_4] = x[ifft_4 + iN1_4*441 + iN2_4] + exp_4*temp[ifft_4 + iN2_4*3 + iN1_sum_4] {id=update_4, dep=exp_4}
        """,  # noqa: E501
        [
            lp.GlobalArg(
                name="x", dtype=np.complex128,
                shape=(1323,), for_atomic=False),
            lp.TemporaryVariable(
                name="exp_table",
                dtype=np.complex128,
                shape=(1323,), for_atomic=False,
                address_space=lp.auto,
                read_only=False,
                ),
            lp.TemporaryVariable(
                name="temp",
                dtype=np.complex128,
                shape=(1323,), for_atomic=False,
                address_space=lp.auto,
                read_only=False,
                ),
            lp.TemporaryVariable(
                name="table_idx_0",
                dtype=np.uint32,
                shape=(), for_atomic=False,
                address_space=lp.auto,
                read_only=False,
                ),
            lp.TemporaryVariable(
                name="exp_0",
                dtype=np.complex128,
                shape=(), for_atomic=False,
                address_space=lp.auto,
                read_only=False,
                ),
            lp.TemporaryVariable(
                name="table_idx_1",
                dtype=np.uint32,
                shape=(), for_atomic=False,
                address_space=lp.auto,
                read_only=False,
                ),
            lp.TemporaryVariable(
                name="exp_1",
                dtype=np.complex128,
                shape=(), for_atomic=False,
                address_space=lp.auto,
                read_only=False,
                ),
            lp.TemporaryVariable(
                name="table_idx_2",
                dtype=np.uint32,
                shape=(), for_atomic=False,
                address_space=lp.auto,
                read_only=False,
                ),
            lp.TemporaryVariable(
                name="exp_2",
                dtype=np.complex128,
                shape=(), for_atomic=False,
                address_space=lp.auto,
                read_only=False,
                ),
            lp.TemporaryVariable(
                name="table_idx_3",
                dtype=np.uint32,
                shape=(), for_atomic=False,
                address_space=lp.auto,
                read_only=False,
                ),
            lp.TemporaryVariable(
                name="exp_3",
                dtype=np.complex128,
                shape=(), for_atomic=False,
                address_space=lp.auto,
                read_only=False,
                ),
            lp.TemporaryVariable(
                name="table_idx_4",
                dtype=np.uint32,
                shape=(), for_atomic=False,
                address_space=lp.auto,
                read_only=False,
                ),
            lp.TemporaryVariable(
                name="exp_4",
                dtype=np.complex128,
                shape=(), for_atomic=False,
                address_space=lp.auto,
                read_only=False,
                ),
            ], name="fft")
    loopy_kernel_knl = lp.make_kernel(
        [
            "{ [i] : 0 <= i <= 1322 }",
            "[m] -> { [j_outer, j_inner]"
            " : j_inner >= 0 and -64j_outer <= j_inner <= 63"
            " and j_inner < m - 64j_outer }",
        ],
        """
        [i]: y[j_inner + j_outer*64, i] = fft([i]: y[j_inner + j_outer*64, i])
        """, [
            lp.ValueArg(
                name="m",
                dtype=None),
            lp.GlobalArg(
                name="y", dtype=None,
                shape=lp.auto),
            ],
        iname_slab_increments={"j_outer": (0, 0)})

    # }}}

    loopy_kernel_knl = lp.tag_inames(loopy_kernel_knl, "j_outer:g.0")
    t_unit = lp.merge([fft_knl, loopy_kernel_knl])
    if inline:
        t_unit = lp.inline_callable_kernel(t_unit, "fft")

    # generate code to ensure that we don't emit spurious missing barrier
    print(lp.generate_code_v2(t_unit).device_code())


def test_inlining_w_zero_stride_callee_args(ctx_factory):
    # See https://github.com/inducer/loopy/issues/594
    ctx = ctx_factory()

    times_two = lp.make_function(
        "{[j1, j2, j3]: 0<=j1,j2,j3<1}",
        """
        out[j1, j2, j3] = 2 * inp[j1, j2, j3]
        """,
        name="times_two")

    knl = lp.make_kernel(
        "{[i1, i2, i3]: 0<=i1,i2,i3<1}",
        """
        tmp[0, 0, 0] = 2.0
        [i1,i2,i3]: y[i1,i2,i3] = times_two([i1,i2,i3]: tmp[0,i2,i3])
        """,
        [lp.TemporaryVariable("tmp", strides=(0, 1, 1)), ...])

    ref_knl = lp.merge([knl, times_two])
    knl = lp.inline_callable_kernel(ref_knl, "times_two")
    lp.auto_test_vs_ref(ref_knl, ctx, knl)


@pytest.mark.parametrize("inline", [True, False])
def test_call_kernel_w_preds(ctx_factory, inline):
    ctx = ctx_factory()
    cq = cl.CommandQueue(ctx)

    twice = lp.make_function(
        "{ [i] : 0<=i<10 }",
        """
        x[i] = 2*x[i]
        """, name="twice")

    knl = lp.make_kernel(
        "{[i] : 0<=i<10}",
        """
        for i
            if i >= 5
                x[i, :] = twice(x[i, :])
            end
        end
        """,
        [lp.GlobalArg("x",
                      shape=(10, 10),),
         ...])

    knl = lp.merge([knl, twice])

    if inline:
        knl = lp.inline_callable_kernel(knl, "twice")

    _evt, (out,) = knl(cq, x=np.ones((10, 10)))

    np.testing.assert_allclose(out[:5], 1)
    np.testing.assert_allclose(out[5:], 2)


# {{{ test_inlining_does_not_lose_preambles

class OurPrintfDefiner(ImmutableRecord):
    def __call__(self, *args, **kwargs):
        return (("42", r"#define OURPRINTF printf"),)


@pytest.mark.parametrize("inline", [True, False])
def test_inlining_does_not_lose_preambles(ctx_factory, inline):
    # loopy.git<=95cc206 would miscompile this

    ctx = ctx_factory()
    cq = cl.CommandQueue(ctx)

    callee = lp.make_function(
        "{ : }",
        [
            lp.CInstruction("",
                            r"""
                            MYPRINTF("Foo bar!\n");
                            """, assignees=())
        ],
        preambles=[("1729", r"#define MYPRINTF OURPRINTF")],
        preamble_generators=[OurPrintfDefiner()],
        name="print_foo")

    caller = lp.make_kernel(
        "{ : }",
        """
        print_foo()
        """,
        name="print_foo_caller")

    knl = lp.merge([caller, callee])
    knl = lp.set_options(knl, "write_code")

    if inline:
        knl = lp.inline_callable_kernel(knl, "print_foo")

    # run to make sure there is no compilation error
    knl(cq)

# }}}


@pytest.mark.parametrize("inline", [True, False])
def test_c_instruction_in_callee(ctx_factory, inline):
    from loopy.symbolic import parse

    ctx = ctx_factory()
    cq = cl.CommandQueue(ctx)
    rng = np.random.default_rng(seed=42)

    knl = lp.make_function(
        "{[i]: 0<=i<10}",
        [lp.CInstruction(iname_exprs=("i", "i"),
                         code="break;",
                         predicates={parse("i >= n")},
                         id="break",),
         lp.Assignment("out_callee", "i", depends_on=frozenset(["break"]))
         ],
        [lp.ValueArg("n", dtype="int32"), ...],
        name="circuit_breaker")

    t_unit = lp.make_kernel(
        "{ : }",
        """
        []: result[0] = circuit_breaker(N)
        """,
        [lp.ValueArg("N", dtype="int32"), ...],)

    t_unit = lp.merge([t_unit, knl])

    if inline:
        t_unit = lp.inline_callable_kernel(t_unit, "circuit_breaker")

    n = rng.integers(3, 8)
    _, (out,) = t_unit(cq, N=n)

    assert out.get() == (n-1)


def test_global_temp_var_with_base_storage(ctx_factory):
    ctx = ctx_factory()
    cq = cl.CommandQueue(ctx)

    knl = lp.make_kernel(
        "{[i, i2] : 0<=i,i2<3}",
        """
        a[i] = 5
        b[i] = a[i] + 1
        ... gbarrier
        c[i2] = b[i2] + 2
        d[i2] = c[i2] + 3
        """, [
            lp.TemporaryVariable("a", dtype=np.int32, shape=(3,),
                address_space=lp.AddressSpace.GLOBAL, base_storage="bs"),
            lp.TemporaryVariable("b", dtype=np.int32, shape=(3,),
                address_space=lp.AddressSpace.GLOBAL, base_storage="bs"),
            lp.TemporaryVariable("c", dtype=np.int32, shape=(3,),
                address_space=lp.AddressSpace.GLOBAL, base_storage="bs"),
            ...
        ],
        seq_dependencies=True)

    knl = lp.allocate_temporaries_for_base_storage(knl, aliased=False)

    cl_prg = cl.Program(ctx, lp.generate_code_v2(knl).device_code()).build()
    assert [knl.num_args for knl in cl_prg.all_kernels()] == [1, 2]

    _evt, (d,) = knl(cq)
    assert (d.get() == 5 + 1 + 2 + 3).all()


def test_inline_deps(ctx_factory):
    # https://github.com/inducer/loopy/issues/564

    ctx = ctx_factory()
    cq = cl.CommandQueue(ctx)

    child_knl = lp.make_function(
        "{[i]:0<=i < 4}",
        """
        a[i] = b[i]
        """, name="func")

    parent_knl = lp.make_kernel(
        "{[i]: 0<=i<4}",
        """
        <> b[i] = i  {id=init,dup=i}
        [i]: a[i] = func([i]: b[i]) {dep=init}
        """)

    prg = lp.merge([parent_knl, child_knl])
    inlined = lp.inline_callable_kernel(prg, "func")

    from loopy.kernel.creation import apply_single_writer_dependency_heuristic
    apply_single_writer_dependency_heuristic(inlined, error_if_used=True)

    _evt, (a_dev,) = inlined(cq)

    assert np.array_equal(a_dev.get(), np.arange(4))


def test_inline_stride():
    # https://github.com/inducer/loopy/issues/728
    child_knl = lp.make_function(
            [],
            """
            g[0] = 2*e[0] + 3*f[0] {id=a}
            g[1] = 2*e[1] + 3*f[1] {dep=a}
            """, name="linear_combo")
    parent_knl = lp.make_kernel(
            ["{[j]:0<=j<n}", "{[i]:0<=i<n}"],
            """
            [i]: z[i, j] = linear_combo([i]: x[i, j], [i]: y[i,j])
            """,
            kernel_data=[
                lp.GlobalArg(
                    name="x, y, z",
                    dtype=np.float64,
                    shape=("n", "n")),
                ...],
            assumptions="n>=2",
            )
    knl = lp.merge([parent_knl, child_knl])
    knl = lp.inline_callable_kernel(knl, "linear_combo")
    lp.generate_code_v2(knl).device_code()


def test_inline_predicate():
    # https://github.com/inducer/loopy/issues/739
    twice = lp.make_function(
        "{[i]: 0<=i<10}",
        """
        y[i] = 2*x[i]
        """,
        name="twice")

    knl = lp.make_kernel(
        "{[i,j]: 0<=i<10 and 0<=j<10}",
        """
        <> y[i] = 0  {id=y,dup=i}
        <> x[i] = 1  {id=x,dup=i}
        for j
            <> a = (j%2 == 0)
            y[i] = i                         {dep=y,if=a,id=y2}
            [i]: z[i, j] = twice([i]: x[i])  {if=a,dep=y}
        end
        """)

    knl = lp.add_dtypes(knl, {"z": np.float64})
    knl = lp.merge([knl, twice])
    knl = lp.inline_callable_kernel(knl, "twice")
    code = lp.generate_code_v2(knl).device_code()
    assert code.count("if (a)") == 1


def test_subarray_ref_with_repeated_indices(ctx_factory):
    # https://github.com/inducer/loopy/pull/735#discussion_r1071690388

    ctx = ctx_factory()
    cq = cl.CommandQueue(ctx)
    child_knl = lp.make_function(
            ["{[i]: 0<=i<10}"],
            """
            g[i] = 1
            """, name="ones")

    parent_knl = lp.make_kernel(
            ["{[i]:0<=i<10}", "{[j]: 0<=j<10}"],
            """
            z[i, j] = 0                {id = a}
            [i]: z[i, i] = ones()  {dep=a,dup=i}
            """,
            kernel_data=[
                lp.GlobalArg(
                    name="z",
                    dtype=np.float64,
                    is_input=False,
                    shape=(10, 10)),
                ...],
            )
    knl = lp.merge([parent_knl, child_knl])
    knl = lp.inline_callable_kernel(knl, "ones")
    _evt, (z_dev,) = knl(cq)
    assert np.allclose(z_dev.get(), np.eye(10))


def test_inline_constant_access():
    child_knl = lp.make_function(
            [],
            """
            g[0] = 2*e[0] + 3*f[0] {id=a}
            g[1] = 2*e[1] + 3*f[1] {dep=a}
            """, name="linear_combo")
    parent_knl = lp.make_kernel(
            ["{[j]:0<=j<n}", "{[i]:0<=i<n}"],
            """
            [i]: z[i, j] = linear_combo([i]: x[i, j], [i]: y[i,j])
            """,
            kernel_data=[
                lp.GlobalArg(
                    name="x, y, z",
                    dtype=np.float64,
                    shape=(3, "n")),
                ...],
            )
    knl = lp.merge([parent_knl, child_knl])
    knl = lp.inline_callable_kernel(knl, "linear_combo")
    knl = lp.tag_array_axes(knl, ["x", "y", "z"], "sep,C")
    lp.generate_code_v2(knl).device_code()


if __name__ == "__main__":
    import sys
    if len(sys.argv) > 1:
        exec(sys.argv[1])
    else:
        from pytest import main
        main([__file__])

# vim: foldmethod=marker



from __future__ import annotations


__copyright__ = "Copyright (C) 2012 Andreas Kloeckner"

__license__ = """
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""


import islpy as isl

from loopy.kernel import LoopKernel
from loopy.symbolic import RuleAwareSubstitutionMapper, SubstitutionRuleMappingContext
from loopy.translation_unit import for_each_kernel


__doc__ = """

.. currentmodule:: loopy

.. autofunction:: fix_parameters

.. autofunction:: assume
"""


# {{{ assume

@for_each_kernel
def assume(kernel, assumptions):
    """Include an assumption about :ref:`domain-parameters` in the kernel, e.g.
    `n mod 4 = 0`.

    :arg assumptions: a :class:`islpy.BasicSet` or a string representation of
        the assumptions in :ref:`isl-syntax`.
    """
    if isinstance(assumptions, str):
        assumptions_set_str = "[%s] -> { : %s}" \
                % (",".join(s for s in kernel.outer_params()),
                    assumptions)
        assumptions = isl.BasicSet.read_from_str(kernel.domains[0].get_ctx(),
                assumptions_set_str)

    if not isinstance(assumptions, isl.BasicSet):
        raise TypeError("'assumptions' must be a BasicSet or a string")

    old_assumptions, new_assumptions = isl.align_two(kernel.assumptions, assumptions)

    return kernel.copy(
            assumptions=old_assumptions.params() & new_assumptions.params())

# }}}


# {{{ fix_parameter

def _fix_parameter(kernel, name, value, within=None):
    def process_set(s):
        var_dict = s.get_var_dict()

        try:
            dt, idx = var_dict[name]
        except KeyError:
            return s

        value_aff = isl.Aff.zero_on_domain(s.space) + value

        from loopy.isl_helpers import iname_rel_aff
        name_equal_value_aff = iname_rel_aff(s.space, name, "==", value_aff)

        s = (s
                .add_constraint(
                    isl.Constraint.equality_from_aff(name_equal_value_aff))
                .project_out(dt, idx, 1))

        return s

    new_domains = [process_set(dom) for dom in kernel.domains]

    from pymbolic.mapper.substitutor import make_subst_func
    subst_func = make_subst_func({name: value})

    from loopy.symbolic import PartialEvaluationMapper, SubstitutionMapper
    subst_map = SubstitutionMapper(subst_func)
    ev_map = PartialEvaluationMapper()

    def map_expr(expr):
        return ev_map(subst_map(expr))

    from loopy.kernel.array import ArrayBase
    new_args = []
    for arg in kernel.args:
        if arg.name == name:
            # remove from argument list
            continue

        if not isinstance(arg, ArrayBase):
            new_args.append(arg)
        else:
            new_args.append(arg.map_exprs(map_expr))

    new_temp_vars = {}
    for tv in kernel.temporary_variables.values():
        new_temp_vars[tv.name] = tv.map_exprs(map_expr)

    from loopy.match import parse_stack_match
    within = parse_stack_match(within)

    rule_mapping_context = SubstitutionRuleMappingContext(
            kernel.substitutions, kernel.get_var_name_generator())
    esubst_map = RuleAwareSubstitutionMapper(
            rule_mapping_context, subst_func, within=within)
    return (
            rule_mapping_context.finish_kernel(
                esubst_map.map_kernel(kernel, within=within,
                    # overwritten below, no need to map
                    map_tvs=False, map_args=False))
            .copy(
                domains=new_domains,
                args=new_args,
                temporary_variables=new_temp_vars,
                assumptions=process_set(kernel.assumptions),
                ))


@for_each_kernel
def fix_parameters(kernel, **value_dict):
    """Fix the values of the arguments to specific constants.

    *value_dict* consists of *name*/*value* pairs, where *name* will be fixed
    to be *value*. *name* may refer to :ref:`domain-parameters` or
    :ref:`arguments`.
    """
    assert isinstance(kernel, LoopKernel)

    within = value_dict.pop("within", None)

    for name, value in value_dict.items():
        kernel = _fix_parameter(kernel, name, value, within)

    return kernel

# }}}

# vim: foldmethod=marker



from __future__ import annotations


__copyright__ = "Copyright (C) 2012 Andreas Kloeckner"

__license__ = """
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""

import logging
from typing import TYPE_CHECKING, Iterable, List, TypeVar, cast


logger = logging.getLogger(__name__)

from functools import partial

import numpy as np
from constantdict import constantdict

from pytools import ProcessLogger

from loopy.diagnostic import (
    LoopyAdvisory,
    LoopyError,
    WriteRaceConditionWarning,
    warn_with_kernel,
)
from loopy.kernel.data import (
    ArrayArg,
    KernelArgument,
    ValueArg,
    _ArraySeparationInfo,
    auto,
    filter_iname_tags_by_type,
)
from loopy.kernel.function_interface import (
    ArgDescriptor,
    CallableKernel,
    ScalarCallable,
)

# from loopy.transform.iname import remove_any_newly_unused_inames
from loopy.kernel.instruction import (
    CallInstruction,
    CInstruction,
    MultiAssignmentBase,
    _DataObliviousInstruction,
)
from loopy.symbolic import RuleAwareIdentityMapper
from loopy.tools import memoize_on_disk
from loopy.translation_unit import TranslationUnit, for_each_kernel

# for the benefit of loopy.statistics, for now
from loopy.type_inference import infer_unknown_types


if TYPE_CHECKING:
    from loopy.kernel import LoopKernel
    from loopy.kernel.array import ArrayDimImplementationTag
    from loopy.typing import Expression


# {{{ check for writes to predicates

def check_for_writes_to_predicates(kernel):
    from loopy.symbolic import get_dependencies
    for insn in kernel.instructions:
        pred_vars = (
                frozenset.union(
                    *(get_dependencies(pred) for pred in insn.predicates))
                if insn.predicates else frozenset())
        written_pred_vars = frozenset(insn.assignee_var_names()) & pred_vars
        if written_pred_vars:
            raise LoopyError("In instruction '%s': may not write to "
                    "variable(s) '%s' involved in the instruction's predicates"
                    % (insn.id, ", ".join(written_pred_vars)))

# }}}


# {{{ check reduction iname uniqueness

def check_reduction_iname_uniqueness(kernel):
    iname_to_reduction_count = {}
    iname_to_nonsimultaneous_reduction_count = {}

    def map_reduction(expr, rec):
        rec(expr.expr)

        for iname in expr.inames:
            iname_to_reduction_count[iname] = (
                    iname_to_reduction_count.get(iname, 0) + 1)
            if not expr.allow_simultaneous:
                iname_to_nonsimultaneous_reduction_count[iname] = (
                        iname_to_nonsimultaneous_reduction_count.get(iname, 0) + 1)

        return expr

    from loopy.symbolic import ReductionCallbackMapper
    cb_mapper = ReductionCallbackMapper(map_reduction)

    for insn in kernel.instructions:
        insn.with_transformed_expressions(cb_mapper)

    for iname, count in iname_to_reduction_count.items():
        nonsimul_count = iname_to_nonsimultaneous_reduction_count.get(iname, 0)

        if nonsimul_count and count > 1:
            raise LoopyError("iname '%s' used in more than one reduction. "
                    "(%d of them, to be precise.) "
                    "Since this usage can easily cause loop scheduling "
                    "problems, this is prohibited by default. "
                    "Use loopy.make_reduction_inames_unique() to fix this. "
                    "If you are sure that this is OK, write the reduction "
                    "as 'simul_reduce(...)' instead of 'reduce(...)'"
                    % (iname, count))

# }}}


# {{{ make_arrays_for_sep_arrays

T = TypeVar("T")


def _remove_at_indices(
        indices: frozenset[int], values: Iterable[T] | None
        ) -> tuple[T, ...] | None:
    """
    Assumes *indices* is sorted.
    """
    if values is None:
        return values

    return tuple(val for i, val in enumerate(values) if i not in indices)


@for_each_kernel
def make_arrays_for_sep_arrays(kernel: LoopKernel) -> LoopKernel:
    from loopy.kernel.array import SeparateArrayArrayDimTag
    new_args = []

    vng = kernel.get_var_name_generator()
    made_changes = False

    # {{{ rewrite arguments

    for arg in kernel.args:
        if not isinstance(arg, ArrayArg) or arg.dim_tags is None:
            new_args.append(arg)
            continue

        sep_axis_indices = [
                i for i, dim_tag in enumerate(arg.dim_tags)
                if isinstance(dim_tag, SeparateArrayArrayDimTag)]

        if not sep_axis_indices or arg._separation_info:
            new_args.append(arg)
            continue

        made_changes = True

        sep_axis_indices_set = frozenset(sep_axis_indices)

        assert isinstance(arg.shape, tuple)
        new_shape: tuple[Expression, ...] | None = \
                _remove_at_indices(sep_axis_indices_set, arg.shape)
        new_dim_tags: tuple[ArrayDimImplementationTag, ...] | None = \
                _remove_at_indices(sep_axis_indices_set, arg.dim_tags)
        new_dim_names: tuple[str | None, ...] | None = \
                _remove_at_indices(sep_axis_indices_set, arg.dim_names)

        sep_shape: list[Expression] = [arg.shape[i] for i in sep_axis_indices]
        for i, sep_shape_i in enumerate(sep_shape):
            if not isinstance(sep_shape_i, (int, np.integer)):
                raise LoopyError(
                        f"Axis {sep_axis_indices[i]+1} (1-based) of "
                        f"argument '{arg.name}' is tagged 'sep', but "
                        "does not have constant length.")

        sep_info = _ArraySeparationInfo(
                sep_axis_indices_set=sep_axis_indices_set,
                subarray_names=constantdict({
                    ind: vng(f"{arg.name}_s{'_'.join(str(i) for i in ind)}")
                    for ind in np.ndindex(*cast("List[int]", sep_shape))}))

        new_args.append(arg.copy(_separation_info=sep_info))

        for _ind, san in sorted(sep_info.subarray_names.items()):
            new_args.append(
                    arg.copy(
                        name=san,
                        shape=new_shape,
                        dim_tags=new_dim_tags,
                        dim_names=new_dim_names))

    # }}}

    if not made_changes:
        return kernel

    kernel = kernel.copy(args=new_args)

    return kernel

# }}}


# {{{ make temporary variables for offsets and strides

def make_args_for_offsets_and_strides(kernel: LoopKernel) -> LoopKernel:
    additional_args: list[KernelArgument] = []

    vng = kernel.get_var_name_generator()

    from pymbolic.primitives import ExpressionNode, Variable

    from loopy.kernel.array import FixedStrideArrayDimTag

    # {{{ process arguments

    new_args = []
    for arg in kernel.args:
        if isinstance(arg, ArrayArg) and not arg._separation_info:
            what = f"offset for argument '{arg.name}'"
            if arg.offset is None:
                pass
            if arg.offset is auto:
                offset_name = vng(arg.name+"_offset")
                additional_args.append(ValueArg(
                        offset_name, kernel.index_dtype))
                arg = arg.copy(offset=offset_name)
            elif isinstance(arg.offset, (int, np.integer, ExpressionNode, str)):
                pass
            else:
                raise LoopyError(f"invalid value of {what}")

            if arg.dim_tags is None:
                new_dim_tags: tuple[ArrayDimImplementationTag, ...] | None \
                        = arg.dim_tags
            else:
                new_dim_tags = ()
                for iaxis, dim_tag in enumerate(arg.dim_tags):
                    if isinstance(dim_tag, FixedStrideArrayDimTag):
                        what = ("axis stride for axis "
                                f"{iaxis+1} (1-based) of '{arg.name}'")
                        if dim_tag.stride is auto:
                            stride_name = vng(f"{arg.name}_stride{iaxis}")
                            dim_tag = dim_tag.copy(stride=Variable(stride_name))
                            additional_args.append(ValueArg(
                                    stride_name, kernel.index_dtype))
                        elif isinstance(
                                dim_tag.stride, (int, np.integer, ExpressionNode)):
                            pass
                        else:
                            raise LoopyError(f"invalid value of {what}")

                    new_dim_tags = (*new_dim_tags, dim_tag)

            arg = arg.copy(dim_tags=new_dim_tags)

        new_args.append(arg)

    # }}}

    if not additional_args:
        return kernel
    else:
        return kernel.copy(args=new_args + additional_args)

# }}}


# {{{ zero_offsets

def zero_offsets_and_strides(kernel: LoopKernel) -> LoopKernel:
    made_changes = False
    from pymbolic.primitives import ExpressionNode

    # {{{ process arguments

    new_args = []
    for arg in kernel.args:
        if isinstance(arg, ArrayArg):
            if arg.offset is None:
                pass
            if arg.offset is auto:
                made_changes = True
                arg = arg.copy(offset=0)
            elif isinstance(arg.offset, (int, np.integer, ExpressionNode, str)):
                from pymbolic.primitives import is_zero
                if not is_zero(arg.offset):
                    raise LoopyError(
                        f"Non-zero offset on argument '{arg.name}' "
                        f"of callable kernel '{kernel.name}. This is not allowed.")
            else:
                raise LoopyError(f"invalid value of offset for '{arg.name}'")

        new_args.append(arg)

    # }}}

    if not made_changes:
        return kernel
    else:
        return kernel.copy(args=new_args)

# }}}


# {{{ decide temporary address space

def _get_compute_inames_tagged(kernel, insn, tag_base):
    return {iname for iname in kernel.insn_inames(insn.id)
               if kernel.iname_tags_of_type(iname, tag_base)}


def _get_assignee_inames_tagged(kernel, insn, tag_base, tv_names):
    return {iname
            for aname, adeps in zip(
                insn.assignee_var_names(),
                insn.assignee_subscript_deps())
            for iname in adeps & kernel.all_inames()
            if aname in tv_names
            if kernel.iname_tags_of_type(iname, tag_base)}


def find_temporary_address_space(kernel):
    logger.debug("%s: find temporary address space" % kernel.name)

    new_temp_vars = {}
    import loopy as lp
    from loopy.kernel.data import AddressSpace, GroupInameTag, LocalInameTagBase

    writers = kernel.writer_map()

    base_storage_to_aliases = {}

    for temp_var in kernel.temporary_variables.values():
        if temp_var.base_storage is not None:
            base_storage_to_aliases.setdefault(
                    temp_var.base_storage, []).append(temp_var.name)

    for temp_var in kernel.temporary_variables.values():
        # Only fill out for variables that do not yet know if they're
        # local. (I.e. those generated by implicit temporary generation.)

        if temp_var.address_space is not lp.auto:
            new_temp_vars[temp_var.name] = temp_var
            continue

        tv_names = (frozenset([temp_var.name])
                | frozenset(base_storage_to_aliases.get(temp_var.base_storage, [])))
        my_writers = writers.get(temp_var.name, frozenset())
        if temp_var.base_storage is not None:
            for alias in base_storage_to_aliases.get(temp_var.base_storage, []):
                my_writers = my_writers | writers.get(alias, frozenset())

        desired_aspace_per_insn = []
        for insn_id in my_writers:
            insn = kernel.id_to_insn[insn_id]

            # A write race will emerge if:
            #
            # - the variable is local
            #   and
            # - the instruction is run across more inames (locally) parallel
            #   than are reflected in the assignee indices.

            locparallel_compute_inames = _get_compute_inames_tagged(
                    kernel, insn, LocalInameTagBase)

            locparallel_assignee_inames = _get_assignee_inames_tagged(
                    kernel, insn, LocalInameTagBase, tv_names)

            grpparallel_compute_inames = _get_compute_inames_tagged(
                    kernel, insn, GroupInameTag)

            grpparallel_assignee_inames = _get_assignee_inames_tagged(
                    kernel, insn, GroupInameTag, temp_var.name)

            assert locparallel_assignee_inames <= locparallel_compute_inames
            assert grpparallel_assignee_inames <= grpparallel_compute_inames

            desired_aspace = AddressSpace.PRIVATE
            for iname_descr, aspace_descr, apin, cpin, aspace in [
                    ("local", "local", locparallel_assignee_inames,
                        locparallel_compute_inames, AddressSpace.LOCAL),
                    ("group", "global", grpparallel_assignee_inames,
                        grpparallel_compute_inames, AddressSpace.GLOBAL),
                    ]:

                if (apin != cpin and bool(apin)):
                    warn_with_kernel(
                            kernel,
                            f"write_race_{aspace_descr}({insn_id})",
                            "instruction '%s' looks invalid: "
                            "it assigns to indices based on %s IDs, but "
                            "its temporary '%s' cannot be made %s because "
                            "a write race across the iname(s) '%s' would emerge. "
                            "(Do you need to add an extra iname to your prefetch?)"
                            % (insn_id, iname_descr, temp_var.name, aspace_descr,
                                ", ".join(cpin - apin)),
                            WriteRaceConditionWarning)

                if (apin == cpin
                        # doesn't want to be in this address space if there
                        # aren't any parallel inames of that kind
                        and bool(cpin)):
                    desired_aspace = max(desired_aspace, aspace)

            desired_aspace_per_insn.append(desired_aspace)

        if not desired_aspace_per_insn:
            warn_with_kernel(kernel, "temp_to_write(%s)" % temp_var.name,
                    "cannot automatically determine address space of '%s'"
                    % temp_var.name, LoopyAdvisory)

            new_temp_vars[temp_var.name] = temp_var
            continue

        overall_aspace = max(desired_aspace_per_insn)

        if not all(iaspace == overall_aspace for iaspace in desired_aspace_per_insn):
            raise LoopyError("not all instructions agree on the "
                    "the desired address space (private/local/global) of  the "
                    "temporary '%s'" % temp_var.name)

        new_temp_vars[temp_var.name] = temp_var.copy(address_space=overall_aspace)

    return kernel.copy(temporary_variables=new_temp_vars)

# }}}


# {{{ realize_ilp

def realize_ilp(kernel):
    logger.debug("%s: add axes to temporaries for ilp" % kernel.name)

    from loopy.kernel.data import IlpBaseTag, VectorizeTag, filter_iname_tags_by_type

    privatizing_inames = frozenset(
        name for name, iname in kernel.inames.items()
        if filter_iname_tags_by_type(iname.tags, (IlpBaseTag, VectorizeTag))
    )

    if not privatizing_inames:
        return kernel

    from loopy.transform.privatize import privatize_temporaries_with_inames
    return privatize_temporaries_with_inames(kernel, privatizing_inames)

# }}}


# {{{ check for loads of atomic variables

def check_atomic_loads(kernel):
    """Find instances of AtomicInit or AtomicUpdate with use of other atomic
    variables to update the atomicity
    """

    logger.debug("%s: check atomic loads" % kernel.name)
    from loopy.kernel.array import ArrayBase
    from loopy.kernel.instruction import Assignment, AtomicLoad
    from loopy.types import AtomicType

    # find atomic variables
    atomicity_candidates = (
            {v.name for v in kernel.temporary_variables.values()
                if isinstance(v.dtype, AtomicType)}
            |
            {v.name for v in kernel.args
                if isinstance(v, ArrayBase)
                and isinstance(v.dtype, AtomicType)})

    new_insns = []
    for insn in kernel.instructions:
        if isinstance(insn, Assignment):
            # look for atomic variables
            atomic_accesses = {a.var_name for a in insn.atomicity}
            accessed_atomic_vars = (insn.dependency_names() & atomicity_candidates)\
                - {insn.assignee_var_names()[0]}
            if not accessed_atomic_vars <= atomic_accesses:
                # if we're missing some
                missed = accessed_atomic_vars - atomic_accesses
                for x in missed:
                    if {x} & atomicity_candidates:
                        insn = insn.copy(
                            atomicity=(*insn.atomicity, AtomicLoad(x)))

        new_insns.append(insn)

    return kernel.copy(instructions=new_insns)

# }}}


# {{{ arg_descr_inference

class ArgDescrInferenceMapper(RuleAwareIdentityMapper):
    """
    Infers :attr:`~loopy.kernel.function_interface.arg_id_to_descr` of
    callables visited in an expression.
    """

    def __init__(self, rule_mapping_context, caller_kernel, clbl_inf_ctx):
        super().__init__(rule_mapping_context)
        self.caller_kernel = caller_kernel
        self.clbl_inf_ctx = clbl_inf_ctx

    def map_call(self, expr, expn_state, assignees=None):
        from pymbolic.mapper.substitutor import make_subst_func
        from pymbolic.primitives import Call, Variable

        from loopy.kernel.array import ArrayBase
        from loopy.kernel.data import ValueArg
        from loopy.kernel.function_interface import (
            ValueArgDescriptor,
            get_arg_descriptor_for_expression,
        )
        from loopy.symbolic import ResolvedFunction, SubstitutionMapper

        if not isinstance(expr.function, ResolvedFunction):
            # ignore if the call is not to a ResolvedFunction
            return super().map_call(expr, expn_state)

        arg_id_to_arg = dict(enumerate(expr.parameters))

        if assignees is not None:
            # If supplied with assignees then this is a CallInstruction
            for i, arg in enumerate(assignees):
                arg_id_to_arg[-i-1] = arg

        arg_id_to_descr = {
            arg_id: get_arg_descriptor_for_expression(self.caller_kernel, arg)
            for arg_id, arg in arg_id_to_arg.items()}
        clbl = self.clbl_inf_ctx[expr.function.name]

        # {{{ translating descriptor expressions to the callable's namespace

        deps_as_params = []
        subst_map = {}

        deps = frozenset().union(*(descr.depends_on()
                                   for descr in arg_id_to_descr.values()))

        assert deps <= self.caller_kernel.all_variable_names()

        for dep in deps:
            caller_arg = self.caller_kernel.arg_dict.get(dep, (self.caller_kernel
                                                               .temporary_variables
                                                               .get(dep)))
            if not (isinstance(caller_arg, ValueArg)
                    or (isinstance(caller_arg, ArrayBase)
                        and caller_arg.shape == ())):
                raise NotImplementedError(f"Obtained '{dep}' as a dependency for"
                        f" call '{expr.function.name}' which is not a scalar.")

            clbl, callee_name = clbl.with_added_arg(caller_arg.dtype,
                                                    ValueArgDescriptor())

            subst_map[dep] = Variable(callee_name)
            deps_as_params.append(Variable(dep))

        mapper = SubstitutionMapper(make_subst_func(subst_map))
        arg_id_to_descr = {id_: descr.map_expr(mapper)
                           for id_, descr in arg_id_to_descr.items()}

        # }}}

        # specializing the function according to the parameter description
        new_clbl, self.clbl_inf_ctx = clbl.with_descrs(arg_id_to_descr,
                                                       self.clbl_inf_ctx)

        self.clbl_inf_ctx, new_func_id = (self.clbl_inf_ctx
                                          .with_callable(expr.function.function,
                                                         new_clbl))

        return Call(ResolvedFunction(new_func_id),
                    tuple(self.rec(child, expn_state)
                          for child in expr.parameters)
                    + tuple(deps_as_params))

    def map_call_with_kwargs(self, expr):
        # See https://github.com/inducer/loopy/pull/323
        raise NotImplementedError

    def __call__(self, expr, kernel, insn, assignees=None):
        from loopy.kernel.data import InstructionBase
        from loopy.symbolic import ExpansionState, UncachedIdentityMapper
        assert insn is None or isinstance(insn, InstructionBase)

        return UncachedIdentityMapper.__call__(self, expr,
                ExpansionState(
                    kernel=kernel,
                    instruction=insn,
                    stack=(),
                    arg_context=constantdict()), assignees=assignees)

    def map_kernel(self, kernel):

        new_insns = []

        for insn in kernel.instructions:
            if isinstance(insn, CallInstruction):
                # In call instructions the assignees play an important in
                # determining the arg_id_to_descr
                mapper = partial(self, kernel=kernel, insn=insn,
                        assignees=insn.assignees)
                new_insns.append(insn.with_transformed_expressions(mapper))
            elif isinstance(insn, MultiAssignmentBase):
                mapper = partial(self, kernel=kernel, insn=insn)
                new_insns.append(insn.with_transformed_expressions(mapper))
            elif isinstance(insn, (_DataObliviousInstruction, CInstruction)):
                new_insns.append(insn)
            else:
                raise NotImplementedError("arg_descr_inference for %s instruction" %
                        type(insn))

        return kernel.copy(instructions=new_insns)


def traverse_to_infer_arg_descr(kernel, callables_table):
    """
    Returns a copy of *kernel* with the argument shapes and strides matching for
    resolved functions in the *kernel*. Refer
    :meth:`loopy.kernel.function_interface.InKernelCallable.with_descrs`.

    .. note::

        Initiates a walk starting from *kernel* to all its callee kernels.
    """
    from loopy.symbolic import SubstitutionRuleMappingContext

    rule_mapping_context = SubstitutionRuleMappingContext(
            kernel.substitutions, kernel.get_var_name_generator())

    arg_descr_inf_mapper = ArgDescrInferenceMapper(rule_mapping_context,
            kernel, callables_table)

    descr_inferred_kernel = rule_mapping_context.finish_kernel(
            arg_descr_inf_mapper.map_kernel(kernel))

    return descr_inferred_kernel, arg_descr_inf_mapper.clbl_inf_ctx


def infer_arg_descr(t_unit: TranslationUnit) -> TranslationUnit:
    """
    Returns a copy of *program* with the
    :attr:`loopy.InKernelCallable.arg_id_to_descr` inferred for all the
    callables.
    """
    from loopy import ValueArg, auto
    from loopy.kernel.array import ArrayBase
    from loopy.kernel.function_interface import ArrayArgDescriptor, ValueArgDescriptor
    from loopy.translation_unit import make_clbl_inf_ctx, resolve_callables

    t_unit = resolve_callables(t_unit)

    clbl_inf_ctx = make_clbl_inf_ctx(t_unit.callables_table,
                                     t_unit.entrypoints)

    for e in t_unit.entrypoints:
        def _tuple_or_none(s):
            if isinstance(s, tuple):
                return s
            elif s in [None, auto]:
                return s
            else:
                return s,

        arg_id_to_descr: dict[str, ArgDescriptor] = {}
        for arg in t_unit[e].args:
            if isinstance(arg, ArrayBase):
                if arg.shape not in (None, auto):
                    arg_id_to_descr[arg.name] = ArrayArgDescriptor(
                            _tuple_or_none(arg.shape), arg.address_space,
                            arg.dim_tags)
            elif isinstance(arg, ValueArg):
                arg_id_to_descr[arg.name] = ValueArgDescriptor()
            else:
                raise NotImplementedError()
        new_callable, clbl_inf_ctx = t_unit.callables_table[e].with_descrs(
                arg_id_to_descr, clbl_inf_ctx)
        clbl_inf_ctx, _new_name = clbl_inf_ctx.with_callable(e, new_callable,
                                                            is_entrypoint=True)

    return clbl_inf_ctx.finish_program(t_unit)

# }}}


# {{{  inline_kernels_with_gbarriers

def inline_kernels_with_gbarriers(program):
    from pytools.graph import compute_topological_order

    from loopy.kernel.instruction import BarrierInstruction
    from loopy.kernel.tools import get_call_graph
    from loopy.transform.callable import inline_callable_kernel

    def has_gbarrier(knl):
        return any((isinstance(insn, BarrierInstruction)
                    and insn.synchronization_kind == "global")
                   for insn in knl.instructions)

    call_graph = get_call_graph(program, only_kernel_callables=True)

    # traverse the kernel calls in a reverse topological sort so that barriers
    # are rightly passed to the entrypoints.
    toposort = compute_topological_order(call_graph,
                                         # pass key to have deterministic codegen
                                         key=lambda x: x
                                         )

    for name in toposort[::-1]:
        if has_gbarrier(program[name]):
            program = inline_callable_kernel(program, name)

    return program

# }}}


def filter_reachable_callables(t_unit):
    from loopy.translation_unit import get_reachable_resolved_callable_ids
    reachable_function_ids = get_reachable_resolved_callable_ids(t_unit
                                                                 .callables_table,
                                                                 t_unit.entrypoints)
    new_callables = {name: clbl for name, clbl in t_unit.callables_table.items()
                     if name in (reachable_function_ids | t_unit.entrypoints)}
    return t_unit.copy(callables_table=constantdict(new_callables))


def _preprocess_single_kernel(kernel: LoopKernel, is_entrypoint: bool) -> LoopKernel:
    from loopy.kernel import KernelState

    prepro_logger = ProcessLogger(logger, "%s: preprocess" % kernel.name)

    kernel = make_arrays_for_sep_arrays(kernel)

    if is_entrypoint:
        kernel = make_args_for_offsets_and_strides(kernel)
    else:
        # No need for offsets internally, we can pass arbitrary pointers.
        kernel = zero_offsets_and_strides(kernel)

    from loopy.check import check_identifiers_in_subst_rules
    check_identifiers_in_subst_rules(kernel)

    # {{{ check that there are no l.auto-tagged inames

    from loopy.kernel.data import AutoLocalInameTagBase
    for name, iname in kernel.inames.items():
        if (filter_iname_tags_by_type(iname.tags, AutoLocalInameTagBase)
                 and name in kernel.all_inames()):
            raise LoopyError("kernel with automatically-assigned "
                    "local axes passed to preprocessing")

    # }}}

    # Ordering restriction:
    # Type inference and reduction iname uniqueness don't handle substitutions.
    # Get them out of the way.

    check_for_writes_to_predicates(kernel)
    check_reduction_iname_uniqueness(kernel)

    # Ordering restriction:
    # add_axes_to_temporaries_for_ilp because reduction accumulators
    # need to be duplicated by this.

    kernel = realize_ilp(kernel)

    kernel = find_temporary_address_space(kernel)

    # check for atomic loads, much easier to do here now that the dependencies
    # have been established
    kernel = check_atomic_loads(kernel)

    kernel = kernel.target.preprocess(kernel)

    kernel = kernel.copy(
            state=KernelState.PREPROCESSED)

    prepro_logger.done()

    return kernel


@memoize_on_disk
def preprocess_program(t_unit: TranslationUnit) -> TranslationUnit:

    from loopy.kernel import KernelState
    if t_unit.state >= KernelState.PREPROCESSED:
        return t_unit

    if len([clbl for clbl in t_unit.callables_table.values() if
            isinstance(clbl, CallableKernel)]) == 1:
        t_unit = t_unit.with_entrypoints(",".join(clbl.name for clbl in
            t_unit.callables_table.values() if isinstance(clbl,
                CallableKernel)))

    if not t_unit.entrypoints:
        raise LoopyError("Translation unit did not receive any entrypoints")

    from loopy.translation_unit import resolve_callables
    t_unit = resolve_callables(t_unit)

    t_unit = filter_reachable_callables(t_unit)

    t_unit = infer_unknown_types(t_unit, expect_completion=False)

    from loopy.transform.subst import expand_subst
    t_unit = expand_subst(t_unit)

    from loopy.kernel.creation import apply_single_writer_dependency_heuristic
    t_unit = apply_single_writer_dependency_heuristic(t_unit)

    # Ordering restrictions:
    #
    # - realize_reduction must happen after type inference because it needs
    #   to be able to determine the types of the reduced expressions.
    #
    # - realize_reduction must happen after default dependencies are added
    #   because it manipulates the depends_on field, which could prevent
    #   defaults from being applied.

    from loopy.transform.realize_reduction import realize_reduction
    t_unit = realize_reduction(t_unit, unknown_types_ok=False)

    # {{{ preprocess callable kernels

    # Callable editing restrictions:
    #
    # - should not edit callables_table in :meth:`preprocess_single_kernel`
    #   as we are iterating over it.[1]
    #
    # [1] https://docs.python.org/3/library/stdtypes.html#dictionary-view-objects

    new_callables = {}
    for func_id, in_knl_callable in t_unit.callables_table.items():
        if isinstance(in_knl_callable, CallableKernel):
            new_subkernel = _preprocess_single_kernel(
                    in_knl_callable.subkernel,
                    is_entrypoint=func_id in t_unit.entrypoints)
            in_knl_callable = in_knl_callable.copy(
                    subkernel=new_subkernel)
        elif isinstance(in_knl_callable, ScalarCallable):
            pass
        else:
            raise NotImplementedError("Unknown callable type %s." % (
                type(in_knl_callable).__name__))

        new_callables[func_id] = in_knl_callable

    t_unit = t_unit.copy(callables_table=constantdict(new_callables))

    # }}}

    # infer arg descrs of the callables
    t_unit = infer_arg_descr(t_unit)

    # Ordering restriction:
    # callees with gbarrier in them must be inlined after inferrring arg_descr.
    t_unit = inline_kernels_with_gbarriers(t_unit)

    # {{{ prepare for caching

    # PicklableDtype instances for example need to know the target they're working
    # towards in order to pickle and unpickle them. This is the first pass that
    # uses caching, so we need to be ready to pickle. This means propagating
    # this target information.

    # }}}

    return t_unit


# FIXME: Do we add a deprecation warning?
preprocess_kernel = preprocess_program


# vim: foldmethod=marker



"""
.. autoclass:: Matchable
.. autoclass:: StackMatchComponent
.. autoclass:: StackMatch

.. autofunction:: parse_match

.. autofunction:: parse_stack_match

.. autodata:: ToStackMatchConvertible

Match expressions
^^^^^^^^^^^^^^^^^

.. autoclass:: MatchExpressionBase
.. autoclass:: All
.. autoclass:: And
.. autoclass:: Or
.. autoclass:: Not
.. autoclass:: Id
.. autoclass:: ObjTagged
.. autoclass:: Tagged
.. autoclass:: Writes
.. autoclass:: Reads
.. autoclass:: InKernel
.. autoclass:: Iname

"""

from __future__ import annotations


__copyright__ = "Copyright (C) 2012 Andreas Kloeckner"

__license__ = """
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""

import re
from abc import ABC, abstractmethod
from dataclasses import dataclass
from sys import intern
from typing import TYPE_CHECKING, Protocol, Sequence, Union

from typing_extensions import TypeAlias

from loopy.kernel.instruction import InstructionBase


NoneType = type(None)

from pytools.lex import RE


if TYPE_CHECKING:
    import pytools.tag

    from loopy.kernel import LoopKernel


def re_from_glob(s: str) -> re.Pattern:
    from fnmatch import translate
    return re.compile("^"+translate(s.strip())+"$")


# {{{ parsing

# {{{ lexer data

_and = intern("and")
_or = intern("or")
_not = intern("not")
_openpar = intern("openpar")
_closepar = intern("closepar")

_id = intern("_id")
_tag = intern("_tag")
_writes = intern("_writes")
_reads = intern("_reads")
_in_kernel = intern("_in_kernel")
_iname = intern("_iname")

_whitespace = intern("_whitespace")

# }}}


_LEX_TABLE = [
    (_and, RE(r"and\b")),
    (_or, RE(r"or\b")),
    (_not, RE(r"not\b")),
    (_openpar, RE(r"\(")),
    (_closepar, RE(r"\)")),

    # TERMINALS
    (_id, RE(r"id:([\w?*]+)")),
    (_tag, RE(r"tag:([\w?*]+)")),
    (_writes, RE(r"writes:([\w?*]+)")),
    (_reads, RE(r"reads:([\w?*]+)")),
    (_in_kernel, RE(r"in_kernel:([\w?*]+)")),
    (_iname, RE(r"iname:([\w?*]+)")),

    (_whitespace, RE("[ \t]+")),
    ]


_TERMINALS = ([_id, _tag, _writes, _reads, _in_kernel, _iname])

# {{{ operator precedence

_PREC_OR = 10
_PREC_AND = 20
_PREC_NOT = 30

# }}}

# }}}


# {{{ match expression

class Matchable(Protocol):
    """
    .. attribute:: tags
    """
    @property
    def tags(self) -> frozenset[pytools.tag.Tag]:
        ...


class MatchExpressionBase(ABC):
    @abstractmethod
    def __call__(self, kernel: LoopKernel, matchable: Matchable) -> bool:
        raise NotImplementedError

    def __ne__(self, other):
        return not self.__eq__(other)

    def __and__(self, other):
        return And((self, other))

    def __or__(self, other):
        return Or((self, other))

    def __inv__(self):
        return Not(self)


class All(MatchExpressionBase):
    def __call__(self, kernel: LoopKernel, matchable: Matchable) -> bool:
        return True

    def __str__(self):
        return "all"

    def __repr__(self):
        return "%s()" % (type(self).__name__)

    def update_persistent_hash(self, key_hash, key_builder):
        key_builder.rec(key_hash, "all_match_expr")

    def __eq__(self, other):
        return type(self) is type(other)

    def __hash__(self):
        return hash(type(self))


@dataclass(frozen=True, eq=True)
class MultiChildMatchExpressionBase(MatchExpressionBase):
    children: Sequence[MatchExpressionBase]

    def __str__(self):
        joiner = " %s " % type(self).__name__.lower()
        return "(%s)" % (joiner.join(str(ch) for ch in self.children))

    def __repr__(self):
        return "{}({})".format(
                type(self).__name__,
                ", ".join(repr(ch) for ch in self.children))


class And(MultiChildMatchExpressionBase):
    def __call__(self, kernel: LoopKernel, matchable: Matchable) -> bool:
        return all(ch(kernel, matchable) for ch in self.children)


class Or(MultiChildMatchExpressionBase):
    def __call__(self, kernel: LoopKernel, matchable: Matchable) -> bool:
        return any(ch(kernel, matchable) for ch in self.children)


@dataclass(frozen=True, eq=True)
class Not(MatchExpressionBase):
    child: MatchExpressionBase

    def __call__(self, kernel: LoopKernel, matchable: Matchable) -> bool:
        return not self.child(kernel, matchable)

    def __str__(self):
        return "(not %s)" % str(self.child)

    def __repr__(self):
        return "{}({!r})".format(type(self).__name__, self.child)


@dataclass(frozen=True, eq=True)
class ObjTagged(MatchExpressionBase):
    """Match if the object is tagged with a given :class:`~pytools.tag.Tag`.

    .. note::

        These instance-based tags will, in the not-too-distant future, replace
        the string-based tags matched by :class:`Tagged`.
    """
    tag: pytools.tag.Tag

    def __call__(self, kernel: LoopKernel, matchable: Matchable) -> bool:
        return self.tag in matchable.tags


class GlobMatchExpressionBase(MatchExpressionBase):
    def __init__(self, glob: str) -> None:
        self.glob = glob

        import re
        from fnmatch import translate
        self.re = re.compile("^"+translate(glob.strip())+"$")

    def __str__(self):
        descr = type(self).__name__
        return descr.lower() + ":" + self.glob

    def __repr__(self):
        return "{}({!r})".format(type(self).__name__, self. glob)

    def update_persistent_hash(self, key_hash, key_builder):
        key_builder.rec(key_hash, type(self).__name__)
        key_builder.rec(key_hash, self.glob)

    def __eq__(self, other):
        return (type(self) is type(other)
                and self.glob == other.glob)

    def __hash__(self):
        return hash((type(self), self.glob))


class Id(GlobMatchExpressionBase):
    def __call__(self, kernel, matchable):
        return self.re.match(matchable.id)


class Tagged(GlobMatchExpressionBase):
    """Match a string-based tagged using a glob expression.

    .. note::

        These string-based tags will, in the not-too-distant future, be replace
        by instance-based tags matched by :class:`ObjTagged`.
    """

    def __call__(self, kernel: LoopKernel, matchable: Matchable) -> bool:
        from loopy.kernel.instruction import LegacyStringInstructionTag
        if matchable.tags:
            return any(
                    self.re.match(tag.value)
                    if isinstance(tag, LegacyStringInstructionTag)
                    else

                    False

                    for tag in matchable.tags)
        else:
            return False


class Writes(GlobMatchExpressionBase):
    def __call__(self, kernel: LoopKernel, matchable: Matchable) -> bool:
        if not isinstance(matchable, InstructionBase):
            return False
        return any(self.re.match(name)
                for name in matchable.assignee_var_names())


class Reads(GlobMatchExpressionBase):
    def __call__(self, kernel: LoopKernel, matchable: Matchable) -> bool:
        if not isinstance(matchable, InstructionBase):
            return False
        return any(self.re.match(name)
                for name in matchable.read_dependency_names())


class InKernel(GlobMatchExpressionBase):
    def __call__(self, kernel, matchable):
        return self.re.match(kernel.name)


class Iname(GlobMatchExpressionBase):
    def __call__(self, kernel: LoopKernel, matchable: Matchable) -> bool:
        if not isinstance(matchable, InstructionBase):
            return False

        return any(self.re.match(name)
                for name in matchable.within_inames)

# }}}


# {{{ parser

def parse_match(expr):
    """Syntax examples::

    * ``id:yoink and writes:a_temp``
    * ``id:yoink and (not writes:a_temp or tag:input)``
    """
    if not expr:
        return All()

    def parse_terminal(pstate):
        next_tag = pstate.next_tag()
        if next_tag is _id:
            result = Id(pstate.next_match_obj().group(1))
            pstate.advance()
            return result
        elif next_tag is _tag:
            result = Tagged(pstate.next_match_obj().group(1))
            pstate.advance()
            return result
        elif next_tag is _writes:
            result = Writes(pstate.next_match_obj().group(1))
            pstate.advance()
            return result
        elif next_tag is _reads:
            result = Reads(pstate.next_match_obj().group(1))
            pstate.advance()
            return result
        elif next_tag is _in_kernel:
            result = InKernel(pstate.next_match_obj().group(1))
            pstate.advance()
            return result
        elif next_tag is _iname:
            result = Iname(pstate.next_match_obj().group(1))
            pstate.advance()
            return result
        else:
            pstate.expected("terminal")

    def inner_parse(pstate, min_precedence=0):
        pstate.expect_not_end()

        if pstate.is_next(_not):
            pstate.advance()
            left_query = Not(inner_parse(pstate, _PREC_NOT))
        elif pstate.is_next(_openpar):
            pstate.advance()
            left_query = inner_parse(pstate)
            pstate.expect(_closepar)
            pstate.advance()
        else:
            left_query = parse_terminal(pstate)

        did_something = True
        while did_something:
            did_something = False
            if pstate.is_at_end():
                return left_query

            next_tag = pstate.next_tag()

            if next_tag is _and and _PREC_AND > min_precedence:
                pstate.advance()
                left_query = And(
                        (left_query, inner_parse(pstate, _PREC_AND)))
                did_something = True
            elif next_tag is _or and _PREC_OR > min_precedence:
                pstate.advance()
                left_query = Or(
                        (left_query, inner_parse(pstate, _PREC_OR)))
                did_something = True

        return left_query

    if isinstance(expr, MatchExpressionBase):
        return expr

    from pytools.lex import InvalidTokenError, LexIterator, lex
    try:
        pstate = LexIterator(
            [(tag, s, idx, matchobj)
             for (tag, s, idx, matchobj) in lex(_LEX_TABLE, expr,
                 match_objects=True)
             if tag is not _whitespace], expr)
    except InvalidTokenError as e:
        from loopy.diagnostic import LoopyError
        raise LoopyError(
                "invalid match expression: '{match_expr}' ({err_type}: {err_str})"
                .format(
                    match_expr=expr,
                    err_type=type(e).__name__,
                    err_str=str(e))) from e

    if pstate.is_at_end():
        pstate.raise_parse_error("unexpected end of input")

    result = inner_parse(pstate)
    if not pstate.is_at_end():
        pstate.raise_parse_error("leftover input after completed parse")

    return result

# }}}


# {{{ stack match objects

class StackMatchComponent(ABC):
    """
    .. automethod:: __call__
    """

    @abstractmethod
    def __call__(self, kernel: LoopKernel, stack: Sequence[Matchable]) -> bool:
        pass

    def __ne__(self, other):
        return not self.__eq__(other)


class StackAllMatchComponent(StackMatchComponent):
    def __call__(self, kernel: LoopKernel, stack: Sequence[Matchable]) -> bool:
        return True

    def update_persistent_hash(self, key_hash, key_builder):
        key_builder.rec(key_hash, "all_match")

    def __eq__(self, other):
        return type(self) is type(other)


class StackBottomMatchComponent(StackMatchComponent):
    def __call__(self, kernel: LoopKernel, stack: Sequence[Matchable]) -> bool:
        return not stack

    def update_persistent_hash(self, key_hash, key_builder):
        key_builder.rec(key_hash, "bottom_match")

    def __eq__(self, other):
        return type(self) is type(other)


@dataclass(eq=True, frozen=True)
class StackItemMatchComponent(StackMatchComponent):
    match_expr: MatchExpressionBase
    inner_match: StackMatchComponent

    def __call__(self, kernel: LoopKernel, stack: Sequence[Matchable]) -> bool:
        if not stack:
            return False

        outer = stack[0]
        if not self.match_expr(kernel, outer):
            return False

        return self.inner_match(kernel, stack[1:])


@dataclass(eq=True, frozen=True)
class StackWildcardMatchComponent(StackMatchComponent):
    inner_match: StackMatchComponent

    def __call__(self, kernel: LoopKernel, stack: Sequence[Matchable]) -> bool:
        for i in range(0, len(stack)):
            if self.inner_match(kernel, stack[i:]):
                return True

        return False

# }}}


# {{{ stack matcher

@dataclass(eq=True, frozen=True)
class RuleInvocationMatchable:
    id: str
    tags: frozenset[pytools.tag.Tag]

    def write_dependency_names(self):
        raise TypeError("writes: query may not be applied to rule invocations")

    def read_dependency_names(self):
        raise TypeError("reads: query may not be applied to rule invocations")

    def inames(self, kernel):
        raise TypeError("inames: query may not be applied to rule invocations")


@dataclass(eq=True, frozen=True)
class StackMatch:
    """
    .. automethod:: __call__
    """

    root_component: StackMatchComponent

    def __call__(
            self, kernel: LoopKernel, insn: InstructionBase,
            rule_stack: Sequence[tuple[str, frozenset[pytools.tag.Tag]]]) -> bool:
        """
        :arg rule_stack: a tuple of (name, tags) rule invocation, outermost first
        """
        stack_of_matchables: list[Matchable] = [insn]
        for id, tags in rule_stack:
            stack_of_matchables.append(RuleInvocationMatchable(id, tags))

        return self.root_component(kernel, stack_of_matchables)

# }}}


# {{{ stack match parsing

ToStackMatchConvertible: TypeAlias = Union[StackMatch, str, None]


def parse_stack_match(smatch: ToStackMatchConvertible) -> StackMatch:
    """Syntax example::

        ... > outer > ... > next > innermost $
        insn > next
        insn > ... > next > innermost $

    ``...`` matches an arbitrary number of intervening stack levels.

    Each of the entries is a match expression as understood by
    :func:`parse_match`.
    """

    if isinstance(smatch, StackMatch):
        return smatch
    if isinstance(smatch, MatchExpressionBase):
        return StackMatch(
                StackItemMatchComponent(
                    smatch, StackAllMatchComponent()))

    if smatch is None:
        return StackMatch(StackAllMatchComponent())

    smatch = smatch.strip()

    match: StackMatchComponent = StackAllMatchComponent()
    if smatch[-1] == "$":
        match = StackBottomMatchComponent()
        smatch = smatch[:-1]

    smatch = smatch.strip()

    components = smatch.split(">")

    for comp in components[::-1]:
        comp = comp.strip()
        if comp == "...":
            match = StackWildcardMatchComponent(match)
        else:
            match = StackItemMatchComponent(parse_match(comp), match)

    return StackMatch(match)

# }}}


# vim: foldmethod=marker



import numpy as np

import pyopencl as cl

import loopy as lp
from loopy.version import LOOPY_USE_LANGUAGE_VERSION_2018_2  # noqa: F401


cl_ctx = cl.create_some_context()

knl = lp.make_kernel(
    "{[ictr,itgt,idim]: "
    "0<=itgt<ntargets "
    "and 0<=ictr<ncenters "
    "and 0<=idim<ambient_dim}",

    """
    for itgt
        for ictr
            <> dist_sq = sum(idim,
                    (tgt[idim,itgt] - center[idim,ictr])**2)
            <> in_disk = dist_sq < (radius[ictr]*1.05)**2
            <> matches = (
                    (in_disk
                        and qbx_forced_limit == 0)
                    or (in_disk
                            and qbx_forced_limit != 0
                            and qbx_forced_limit * center_side[ictr] > 0)
                    )

            <> post_dist_sq = dist_sq if matches else HUGE
        end
        <> min_dist_sq, <> min_ictr = argmin(ictr, ictr, post_dist_sq)

        tgt_to_qbx_center[itgt] = min_ictr if min_dist_sq < HUGE else -1
    end
    """)

knl = lp.fix_parameters(knl, ambient_dim=2)
knl = lp.add_and_infer_dtypes(knl, {
        "tgt,center,radius,HUGE": np.float32,
        "center_side,qbx_forced_limit": np.int32,
        })

lp.auto_test_vs_ref(knl, cl_ctx, knl, parameters={
        "HUGE": 1e20, "ncenters": 200, "ntargets": 300,
        "qbx_forced_limit": 1})



import numpy as np

import loopy as lp
from loopy.version import LOOPY_USE_LANGUAGE_VERSION_2018_2  # noqa: F401


k = lp.make_kernel([
    "{ [i] : 0 <= i < m }",
    "{ [j] : 0 <= j < length }"],
    """
    for i
        <> rowstart = rowstarts[i]
        <> rowend = rowstarts[i+1]
        <> length = rowend - rowstart
        y[i] = sum(j, values[rowstart+j] * x[colindices[rowstart + j]])
    end
    """, name="spmv")

k = lp.add_and_infer_dtypes(k, {
    "values,x": np.float64, "rowstarts,colindices": k["spmv"].index_dtype
    })
print(lp.generate_code_v2(k).device_code())
